﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 * and the BEACON Center for the Study of Evolution in Action.
 * 
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using HeuristicLab.Common;

namespace HeuristicLab.Algorithms.DataAnalysis {
  // loss function for the weighted absolute error
  public class AbsoluteErrorLoss : ILossFunction {
    public double GetLoss(IEnumerable<double> target, IEnumerable<double> pred, IEnumerable<double> weight) {
      var targetEnum = target.GetEnumerator();
      var predEnum = pred.GetEnumerator();
      var weightEnum = weight.GetEnumerator();

      double s = 0;
      while (targetEnum.MoveNext() & predEnum.MoveNext() & weightEnum.MoveNext()) {
        double res = targetEnum.Current - predEnum.Current;
        s += weightEnum.Current * Math.Abs(res);  // w * |res|
      }
      if (targetEnum.MoveNext() | predEnum.MoveNext() | weightEnum.MoveNext())
        throw new ArgumentException("target, pred and weight have differing lengths");

      return s;
    }

    public IEnumerable<double> GetLossGradient(IEnumerable<double> target, IEnumerable<double> pred, IEnumerable<double> weight) {
      var targetEnum = target.GetEnumerator();
      var predEnum = pred.GetEnumerator();
      var weightEnum = weight.GetEnumerator();

      while (targetEnum.MoveNext() & predEnum.MoveNext() & weightEnum.MoveNext()) {
        // dL(y, f(x)) / df(x) = weight * sign(res)
        var res = targetEnum.Current - predEnum.Current;
        if (res > 0) yield return weightEnum.Current;
        else if (res < 0) yield return -weightEnum.Current;
        else yield return 0.0;
      }
      if (targetEnum.MoveNext() | predEnum.MoveNext() | weightEnum.MoveNext())
        throw new ArgumentException("target, pred and weight have differing lengths");
    }

    // return median of residuals
    public LineSearchFunc GetLineSearchFunc(IEnumerable<double> target, IEnumerable<double> pred, IEnumerable<double> weight) {
      var targetArr = target.ToArray();
      var predArr = pred.ToArray();
      var weightArr = weight.ToArray();
      // the optimal constant value that should be added to the predictions is the median of the residuals
      // weights are not supported yet (need to calculate a weighted median)
      Debug.Assert(weightArr.All(w => w.IsAlmost(1.0)));

      if (targetArr.Length != predArr.Length || predArr.Length != weightArr.Length)
        throw new ArgumentException("target, pred and weight have differing lengths");

      // line search for abs error
      LineSearchFunc lineSearch = (idx, startIdx, endIdx) => {
        // Median() is allocating an array anyway
        // It would be possible to pre-allocated an array for the residuals if Median() would allow specification of a sub-range
        int nRows = endIdx - startIdx + 1;
        var res = new double[nRows];
        for (int i = startIdx; i <= endIdx; i++) {
          var row = idx[i];
          res[i - startIdx] = targetArr[row] - predArr[row];
        }
        return res.Median();
      };
      return lineSearch;

    }

    public override string ToString() {
      return "Absolute error loss";
    }
  }
}
