﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 * and the BEACON Center for the Study of Evolution in Action.
 * 
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using HeuristicLab.Common;

namespace HeuristicLab.Algorithms.DataAnalysis {
  // Greedy Function Approximation: A Gradient Boosting Machine (page 9) 
  public class LogisticRegressionLoss : ILossFunction {
    public double GetLoss(IEnumerable<double> target, IEnumerable<double> pred, IEnumerable<double> weight) {
      var targetEnum = target.GetEnumerator();
      var predEnum = pred.GetEnumerator();
      var weightEnum = weight.GetEnumerator();

      double s = 0;
      while (targetEnum.MoveNext() & predEnum.MoveNext() & weightEnum.MoveNext()) {
        Debug.Assert(targetEnum.Current.IsAlmost(0.0) || targetEnum.Current.IsAlmost(1.0), "labels must be 0 or 1 for logistic regression loss");

        var y = targetEnum.Current * 2 - 1; // y in {-1,1}
        s += weightEnum.Current * Math.Log(1 + Math.Exp(-2 * y * predEnum.Current));
      }
      if (targetEnum.MoveNext() | predEnum.MoveNext() | weightEnum.MoveNext())
        throw new ArgumentException("target, pred and weight have different lengths");

      return s;
    }

    public IEnumerable<double> GetLossGradient(IEnumerable<double> target, IEnumerable<double> pred, IEnumerable<double> weight) {
      var targetEnum = target.GetEnumerator();
      var predEnum = pred.GetEnumerator();
      var weightEnum = weight.GetEnumerator();

      while (targetEnum.MoveNext() & predEnum.MoveNext() & weightEnum.MoveNext()) {
        Debug.Assert(targetEnum.Current.IsAlmost(0.0) || targetEnum.Current.IsAlmost(1.0), "labels must be 0 or 1 for logistic regression loss");
        var y = targetEnum.Current * 2 - 1; // y in {-1,1}

        yield return weightEnum.Current * 2 * y / (1 + Math.Exp(2 * y * predEnum.Current));

      }
      if (targetEnum.MoveNext() | predEnum.MoveNext() | weightEnum.MoveNext())
        throw new ArgumentException("target, pred and weight have different lengths");
    }

    public LineSearchFunc GetLineSearchFunc(IEnumerable<double> target, IEnumerable<double> pred, IEnumerable<double> weight) {
      var targetArr = target.ToArray();
      var predArr = pred.ToArray();
      var weightArr = weight.ToArray();
      // weights are not supported yet
      // when weights are supported we need to calculate a weighted median
      Debug.Assert(weightArr.All(w => w.IsAlmost(1.0)));

      if (targetArr.Length != predArr.Length || predArr.Length != weightArr.Length)
        throw new ArgumentException("target, pred and weight have different lengths");

      // "Simple Newton-Raphson step" of eqn. 23 
      LineSearchFunc lineSearch = (idx, startIdx, endIdx) => {
        double sumY = 0.0;
        double sumDiff = 0.0;
        for (int i = startIdx; i <= endIdx; i++) {
          var row = idx[i];
          var y = targetArr[row] * 2 - 1; // y in {-1,1}
          var pseudoResponse = weightArr[row] * 2 * y / (1 + Math.Exp(2 * y * predArr[row]));

          sumY += pseudoResponse;
          sumDiff += Math.Abs(pseudoResponse) * (2 - Math.Abs(pseudoResponse));
        }
        // prevent divByZero
        sumDiff = Math.Max(1E-12, sumDiff);
        return sumY / sumDiff;
      };
      return lineSearch;

    }

    public override string ToString() {
      return "Logistic regression loss";
    }
  }
}
